import logging
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from yamlex.api.splitter import split_yaml
from yamlex.api.util import (
    adjust_root_logger,
    get_default_extension_dir_path,
    get_default_extension_source_dir_path,
    is_manually_created,
    write_file,
    indent,
)
from yamlex.cli.common_flags import (
    no_file_header_flag,
    force_flag,
    verbose_flag,
    quiet_flag,
    dry_run_flag,
    line_length_option,
    remove_comments_flag,
)


logger = logging.getLogger(__name__)


def split(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help="Path to source extension.yaml file.",
            show_default="extension/extension.yaml or src/extension/extension.yaml",
            dir_okay=False,
            file_okay=True,
            exists=True,
            readable=True,
        ),
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help="Path to directory where split YAML source files will be stored.",
            show_default="source or src/source",
            dir_okay=True,
            file_okay=False,
            exists=True,
            writable=True,
        ),
    ] = None,
    line_length: line_length_option = None,
    dry_run: dry_run_flag = False,
    remove_comments: remove_comments_flag = False,
    no_file_header: no_file_header_flag = False,
    force: force_flag = False,
    verbose: verbose_flag = False,
    quiet: quiet_flag = False,
) -> None:
    """
    Split [i]extension.yaml[/i] file into individual components.

    Performs a "best effort" opinionated splitting. This operation does not
    affect the original extension.yaml file specified by --source. Instead,
    it extracts components from it and places them into individual files
    within the --target folder.

    [b]Splitting multiple times[/b]:

    Theoretically, you only need to split once. However, in practice, you
    can do it over and over again (not very well tested). When performing
    consequent splittings, the command overwrites any previously generated
    split files, if they have the 'Generated by yamlex' header within them.
    If the target split part does not have that header, it is considered
    to be manually created and is not overwritten. You can still force
    the overwrite using the --force flag.

    [b]Remove 'Generated with yamlex' header from split files[/b]:

    When splitting, you can choose to not add the 'Generated by yamlex'
    header to the generated files by using the --no-file-header flag.
    """
    adjust_root_logger(verbose, quiet)

    source = source or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"Source file: {source}")

    target = target or get_default_extension_source_dir_path()
    logger.debug(f"Target directory: {target}")

    split_parts = split_yaml(
        source,
        target,
        remove_comments=remove_comments,
    )

    parts_to_skip: list[Path] = []
    for path in split_parts.keys():
        if is_manually_created(path) and not force:
            part = split_parts.pop(path)
            parts_to_skip.append(part)

    if parts_to_skip:
        logger.info((
            "The following split parts will not be written, because "
            "an unmarked file with the same name already exists in the "
            "target path. Unmarked files do not have the "
            "'Generated with yamlex' header, and thus are considered to be "
            "manually created or edited. Use --force to overwrite this "
            "behavior."
        ))
        for p in parts_to_skip:
            logger.info(f"{indent(1)}Skip: {p}")

    if not split_parts:
        logger.info("No part files to be written.")
    else:
        logger.info("The following part files will be written:")
        for path, part in split_parts.items():
            write_file(
                path,
                part,
                add_file_header=not no_file_header,
                line_length=line_length,
                dry_run=dry_run,
                print_to_stdout=False,
            )
            logger.info(f"{indent(1)}Added: {path}")



import logging
import re
import sys
from io import StringIO
from pathlib import Path
from typing import Any, Optional, Union, MutableMapping, MutableSequence

import ruamel.yaml
from ruamel.yaml.comments import CommentedBase, Comment

from yamlex.api.exceptions import (
    NoValidVersionNumber,
    ExtensionDirNotFound,
    FailedToReadVersionFile,
    FailedToWriteVersionFile,
)


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def adjust_root_logger(verbose: bool = False, quiet: bool = False) -> None:
    if quiet:
        logging.root.setLevel(logging.ERROR)
    elif verbose:
        logging.root.setLevel(logging.DEBUG)


def indent(num: int = 0) -> str:
    return "· " * num


def get_default_extension_source_dir_path() -> Path:
    src_dir_path = Path("src")
    if src_dir_path.exists() and src_dir_path.is_dir():
        return src_dir_path / "source"
    else:
        return Path("source")


def get_default_extension_dir_path() -> Path:
    candidates = [
        Path("extension"),
        Path("src/extension"),
    ]

    # Find the first suitable candidate
    for candidate in candidates:
        # Check the candidate directory exists
        if candidate.exists() and candidate.is_dir():
            return candidate
        continue

    # Nothing suitable was found
    logger.error(
        (
            "No default extension directory found. "
            "'extension/' or 'src/extension/' folder must exist "
            "in the current working directory."
        )
    )
    raise ExtensionDirNotFound()


def is_manually_created(path: Path) -> bool:
    if path.exists():
        with open(path, "r") as f:
            content = f.read().lower()
            if "generated by yamlex" not in content:
                True
    return False


def sanitize_file_stem(stem: str) -> Path:
    valid_id = re.sub(r"[^a-z0-9\.]", "_", stem.lower())
    return valid_id


def write_file(
    file_path: Path,
    data: Union[dict, list],
    add_file_header: bool = True,
    line_length: Optional[int] = None,
    dry_run: bool = False,
    print_to_stdout: bool = False,
) -> None:
    # Convert dict to YAML. Dump to string first to add a comment
    stream = StringIO()
    parser.indent(mapping=2, sequence=4, offset=2)
    parser.width = line_length or sys.maxsize
    parser.dump(data, stream)
    text = stream.getvalue()

    # Write a comment to indicate that the file was automatically generated
    header = f"# Generated by yamlex\n\n"

    # Dry run simply means we are not creating the files
    if not dry_run:
        # Make sure the directory we write to exists
        dir = file_path.parent.resolve()
        dir.mkdir(parents=True, exist_ok=True)

        # Write to output file
        with open(file_path, "w") as f:
            if add_file_header:
                f.write(header)
            f.write(text)

    # If dry run mode is active then we might be asked to print to stdout
    elif print_to_stdout:
        if add_file_header:
            print(header, end=None)
        print(text)


def read_version_properties(path: Path, default: Optional[str] = None) -> str:
    try:
        with open(path, "r") as f:
            content = f.read()
            match_ = re.search(r"version=(.*)", content)
            if not match_ and not default:
                raise FailedToReadVersionFile(path)
            return match_.group(1)
    except Exception as e:
        logger.error(f"Failed to open {path}: {e}")
        raise FailedToReadVersionFile(e)


def write_version_properties(path: Path, version: str) -> None:
    try:
        with open(path, "r") as rf:
            lines = rf.readlines()
            content = []
            for l in lines:
                if l.startswith("version"):
                    content.append(f"version={version}\n")
                else:
                    content.append(l)
        with open(path, "w") as wf:
            wf.writelines(content) 
    except Exception as e:
        logger.error(f"Failed to write to {path}: {e}")
        raise FailedToWriteVersionFile(e)


def parse_version(version: str) -> Optional[str]:
    regex = r"^(0|[1-9][0-9]{0,3})(\.(0|[1-9][0-9]{0,3})){0,2}$"
    match_ = re.search(regex, version)
    if not match_:
        return None
    return match_.group()


def bump_version(version: str) -> str:
    version_parts = version.split(".")
    part_int: int | None = None
    part_idx: int | None = None
    for i, part in enumerate(reversed(version_parts)):
        try:
            part_int = int(part)
            part_idx = len(version_parts) - 1 - i
            break
        except:
            continue
    if part_int is None:
        logger.error("Could not find integer part in version")
        raise NoValidVersionNumber()

    part_int += 1
    result = ""
    for i, part in enumerate(version_parts):
        append = part
        if i == part_idx:
            append = part_int
        result += f"{append}."
    result = result.rstrip(".")
    return result


def remove_yaml_comments(
    key: str,
    obj: Any,
    recursive: bool = False,
    level: int = 0,
) -> Any:
    indentation = indent(level)

    if isinstance(obj, CommentedBase):
        if hasattr(obj, Comment.attrib):
            delattr(obj, Comment.attrib)
            logger.debug(f"{indentation}Removed comments from {str(key)}")

    if isinstance(obj, MutableMapping):
        result = {}
        for k, v in obj.items():
            result[k] = remove_yaml_comments(
                k,
                v,
                recursive=True,
                level=level + 1,
            ) if recursive else v
        return result

    if isinstance(obj, MutableSequence):
        result = []
        for i, v in enumerate(obj):
            result.append(
                remove_yaml_comments(
                    i,
                    v,
                    recursive=True,
                    level=level + 1,
                ) if recursive else v
            )
        return result

    return obj



import json
import logging
from pathlib import Path
from typing import Optional

import ruamel.yaml
import typer
from typing_extensions import Annotated

from yamlex.api.differ import diff as diff_data
from yamlex.api.util import adjust_root_logger
from yamlex.cli.common_flags import (
    verbose_flag,
    quiet_flag,
)


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def diff(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help="Path to the source directory or YAML file.",
            show_default=False,
            dir_okay=True,
            file_okay=True,
            exists=True,
            readable=True,
        )
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help="Path to the target directory or YAML file.",
            show_default=False,
            dir_okay=True,
            file_okay=True,
            exists=True,
            readable=True,
        )
    ] = None,
    verbose: verbose_flag = False,
    quiet: quiet_flag = False,
) -> None:
    """
    Compare --source YAML file or directory to --target and print the differences.

    Recursively compares the --source and --target YAML files or directories and
    prints the differences in JSON format. Diff does not care about the
    formatting of the --source and --target, only about the actual content.

    The data from --source will be marked as "old" in the diff and the data
    from --target will be marked as "new".

    Both --source and --target can be a file or a directory.

    Exits with exit code 0 if there is no difference. Otherwise, the exit
    code is 1.
    """
    adjust_root_logger(verbose, quiet)

    logger.debug(f"Source path: {source}")
    logger.debug(f"Target path: {target}")

    # Compare source to target
    difference = diff_data(
        source=source,
        target=target,
    )

    serialized = json.dumps(
        difference,
        indent=2,
        default=str,
    )
    print(serialized)

    if difference:
        raise typer.Exit(1)



from typing import Optional

import typer
from typing_extensions import Annotated


force_flag = Annotated[
    bool,
    typer.Option(
        "--force",
        "-f",
        help="Overwrite target files even if they were created manually.",
    ),
]
no_file_header_flag = Annotated[
    bool,
    typer.Option(
        "--no-file-header",
        "-H",
        help="Do not add the 'generated by yamlex' comment header at the top of the file.",
    ),
]
verbose_flag = Annotated[
    bool,
    typer.Option(
        "--verbose",
        help="Enable verbose output.",
    ),
]
quiet_flag = Annotated[
    bool,
    typer.Option(
        "--quiet",
        help="Disable any informational output. Only errors.",
    ),
]
dry_run_flag = Annotated[
    bool,
    typer.Option(
        "--dry-run",
        help=(
            "Test run without any files being created. "
        ),
    ),
]
remove_comments_flag = Annotated[
    bool,
    typer.Option(
        "--remove-comments",
        help="Remove any YAML comments from the created files."
    ),
]
line_length_option = Annotated[
    Optional[int],
    typer.Option(
        "--line-length",
        help="Maximum line length in the generated extension.yaml.",
        show_default="not limited",
    ),
]


import logging
from pathlib import Path

import ruamel.yaml
from deepdiff import DeepDiff

from yamlex.api.joiner import assemble_recursively
from yamlex.api.util import remove_yaml_comments
from yamlex.api.exceptions import (
    FailedToParseYamlError,
    InvalidPath,
)


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def diff(
    source: Path,
    target: Path,
) -> dict:
    """
    Compare two YAML files recursively and return the differences.

    Returns:
        dict: A dictionary containing the differences between the two files.
    """
    source_data = parse_path(source)
    target_data = parse_path(target)

    differences = DeepDiff(
        source_data,
        target_data,
        ignore_order=True,
        report_repetition=True,
        verbose_level=2,
    )

    return differences


def parse_path(path: Path) -> dict:
    if path.is_file(): 
        try:
            with open(path, "r") as file:
                raw_data: dict = parser.load(file)
        except Exception as e:
            raise FailedToParseYamlError(
                f"Failed to parse YAML in {path}: {e}"
            )
        data = remove_yaml_comments(
            path,
            raw_data,
            recursive=True,
        )
    elif path.is_dir():
        data = assemble_recursively(
            path,
            remove_comments=True,
        )
    else:
        raise InvalidPath(
            f"{path} must be a directory or a file."
        )
    return data



import logging
import sys
from typing import Optional
from typing_extensions import Annotated

import typer
from rich import print

from yamlex.cli.commands.join import join as join_cmd
from yamlex.cli.commands.map import map as map_cmd
from yamlex.cli.commands.split import split as split_cmd
from yamlex.cli.commands.diff import diff as diff_cmd
from yamlex.cli.version import get_version
from yamlex.api.exceptions import (
    YamlexError,
)


logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)


help_text = """
[b]How assembling works:[/b]

- [b]Hierarchy preserved[/b]
  
  Any folder or file within the --source directory is considered to be
  a field within the final extension.yaml. For example, a file called
  [i]name.yaml[/i] will become a field called [i]name:[/i] and its content
  will be put into that field.

  [i]Example 1[/i]:
  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ └── author.yaml                   │ name: John Doe                    │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ author:                           │                                   │
  │   name: John Doe                  │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘
    
- [b]Arrays with '-' symbol[/b]

  The minus sign '-' in front of a file or directory inside a parent
  folder means that the parent folder is an array. Every item within
  that folder will become an array item and the folder itself will
  be added as an array to the final yaml.

  [i]Example 2[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── metrics/                      │                                   │
  │ │   └── -metric_1.yaml            │ key: metric_1                     │
  │ └── +index.yaml                   │ name: com.example.extension.test  │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  │ metrics:                          │                                   │
  │   - key: metric_1                 │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

  [i]Example 3 (same result, using different structure)[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── metrics/                      │                                   │
  │ │   └── -metric_1/                │                                   │
  │ │       └── index.yaml            │ key: metric_1                     │
  │ └── +index.yaml                   │ name: com.example.extension.test  │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  │ metrics:                          │                                   │
  │   - key: metric_1                 │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

- [b]Groupers with '+' symbol[/b]

  If the name of a folder or file starts with a + symbol, then the
  content of that file or folder is added on the same level as the folder
  itself.

  The folder itself is ignored, but [i]everything[/i] inside it is used.
  With the + in front, the folder or file become a "virtual"
  hierarchy, which doesn't add a new nesting level in the final yaml.
  However, anything inside such folder or file will be added on the same
  level as the parent. Almost as if the +folder/ or +file.yaml do not
  exist at all and their content just gets added to the parent.

  [i]Example 4[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── author.yaml                   │ name: John Doe                    │
  │ └── +index.yaml                   │ name: com.example.extension.test  │
  │                                   │ minDynatraceVersion: 1.295.0      │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  │ minDynatraceVersion: 1.295.0      │                                   │
  │ author:                           │                                   │
  │   name: John Doe                  │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

  [i]Example 5[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── author.yaml                   │ name: John Doe                    │
  │ └── +grouper.yaml                 │ name: com.example.extension.test  │
  │                                   │ minDynatraceVersion: 1.295.0      │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  │ minDynatraceVersion: 1.295.0      │                                   │
  │ author:                           │                                   │
  │   name: John Doe                  │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

  [i]Example 6[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── metrics/                      │                                   │
  │ │   └── +group_of_metrics.yaml    │ - key: metric_1                   │
  │ └── +index.yaml                   │ name: com.example.extension.test  │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  │ metrics:                          │                                   │
  │   - key: metric_1                 │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

  [i]Example 7[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── metrics/                      │                                   │
  │ │   └── +group_of_metrics/        │                                   │
  │ │       └── -metric_1.yaml        │ key: metric_1                     │
  │ └── +index.yaml                   │ name: com.example.extension.test  │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  │ metrics:                          │                                   │
  │   - key: metric_1                 │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

- [b]Ignoring with '!' symbol[/b]

  If a folder or a file starts with the exclamation mark symbol '!', then
  it will be ignored, as if it doesn't exist at all, including all its
  content. This is useful for excluding files or folders from the final
  yaml.

  [i]Example 8[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ ├── !metrics/                     │                                   │
  │ │   └── -metric_1.yaml            │ key: metric_1                     │
  │ └── +index.yaml                   │ name: com.example.extension.test  │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ name: com.example.extension.test  │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘

- [b]Scalar files[/b]

  Any plain text file (not yaml) will be added as a scalar value. With
  the formatting preserved. 

  [i]Example 9[/i]:

  ┌───────────────────────────────────┬───────────────────────────────────┐
  │ [i]Source folder structure:[/i]          │ [i]Contents of the file:[/i]             │
  │ source/                           │                                   │
  │ └── vars/                         │                                   │
  │     └── -timeout/                 │                                   │
  │         ├── description.txt       │ Custom timeout.                   │
  │         └── +index.yaml           │ id: timeout                       │
  │                                   │ type: text                        │
  │                                   │ defaultValue: "120"               │
  ├───────────────────────────────────┼───────────────────────────────────┤
  │ [i]Resulting extension.yaml:[/i]         │                                   │
  │ vars:                             │                                   │
  │   - id: timeout                   │                                   │
  │     description: Custom timeout.  │                                   │
  │     type: text                    │                                   │
  │     defaultValue: "120"           │                                   │
  └───────────────────────────────────┴───────────────────────────────────┘
"""


def callback(
    ctx: typer.Context,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            show_default=False,
            is_eager=True,
        ),
    ] = None,
) -> None:
    if version:
        prog_version = get_version()
        print(prog_version)

        raise typer.Exit()
    
    if ctx.invoked_subcommand:
        return


def run() -> None:
    # Set up Typer app.
    app = typer.Typer(
        name="yamlex",
        rich_markup_mode="rich",
        add_completion=False,
        context_settings={"help_option_names": ["-h", "--help"]},
        help=help_text,
        pretty_exceptions_enable=False,
    )

    # Set primary callback. This shows helps or allows to show version.
    app.callback(invoke_without_command=True, no_args_is_help=True)(callback)

    # Declare main commands.
    app.command(name="map")(map_cmd)
    app.command(name="split")(split_cmd)
    app.command(name="join")(join_cmd)
    app.command(name="diff")(diff_cmd)

    # Declare aliases for popular commands.
    app.command(name="j", hidden=True)(join_cmd)
    app.command(name="d", hidden=True)(diff_cmd)

    # Launch Typer app.
    try:
        app()
    except YamlexError as e:
        code = e.code
        name = e.__class__.__name__
        message = f" {e.args[0]}" if e.args else ""
        print(f"Error {code} ({name})!{message}", file=sys.stderr)
        exit(code)



import sys
import logging
import warnings
from pathlib import Path
from typing import Union

import ruamel.yaml
from ruamel.yaml.scalarstring import FoldedScalarString
from ruamel.yaml.comments import (
    CommentedBase,
    CommentedMap,
    CommentedSeq,
    Comment,
)

from .util import remove_yaml_comments, indent as indentation
from .exceptions import (
    InvalidItemWithinArrayDirectoryError,
    UnintendedIndexFileWarning,
    NonTextFileError,
    FailedToParseYamlError,
    IndexFileIsArray,
    DuplicateKey,
)


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def assemble_recursively(
    dir_path: Path,
    keep_formatting: bool = True,
    sort_paths: bool = False,
    dry_run: bool = False,
    remove_comments: bool = False,
    level: int = 0,
) -> Union[dict, list]:
    indent = indentation(level)
    logger.debug(f"{indent}Assembling level: {dir_path}")

    # Get all files in this directory that we will process, but ignore
    # symlinks and paths starting with '!'
    all_paths = [
        p for p in dir_path.glob("*")
        if not (p.is_symlink() or p.name.startswith("!"))
    ]

    # Enable alphabetically sorted keys for fancy users
    if sort_paths:
        all_paths.sort(key=lambda p: p.name)

    # We deal with three types of paths within the directory:
    # 1. Directories
    all_dirs: list[Path] = []
    # 2. YAML files
    all_yamls: dict[str, Path] = {}
    # 3. Non-YAML scalar files, such as SQL queries, DQL, text files, etc.
    scalar_files: dict[str, Path] = {}
    for p in all_paths:
        if p.is_dir():
            all_dirs.append(p)
        elif p.is_file():
            if p.suffix in (".yaml", ".yml"):
                all_yamls[p.stem] = p
            else:
                scalar_files[p.stem] = p

    # All parsed data will be collected into a single data object.
    # However, we don't know in advance, whether we are dealing with
    # a dictionary or a list.
    data: dict[str, Union[dict, list, str, FoldedScalarString]] = {}

    # Parse data from YAML files.
    for yaml_file_name, yaml_file_path in all_yamls.items():
        with open(yaml_file_path, "r") as yaml_file:
            try:
                yaml_file_data = parser.load(yaml_file)
            except Exception as e:
                raise FailedToParseYamlError(
                    f"Failed to parse {yaml_file_path} in {dir_path}. "
                    "Please make sure the YAML syntax is correct. "
                    f"The exact parsing error is: {e}"
                )
            
            if yaml_file_name in data:
                raise DuplicateKey(
                    f"Duplicate key found inside {dir_path}: {yaml_file_name}"
                )
            
            # Remove comments if necessary
            data[yaml_file_name] = remove_yaml_comments(
                yaml_file_name,
                yaml_file_data,
                recursive=True,
                level=level + 1,
            ) if remove_comments else yaml_file_data

    # Load data from scalar files.
    # Example: query.sql file containing a SQL query
    for scalar_file_name, scalar_file_path in scalar_files.items():
        with open(scalar_file_path, "r") as scalar_file:
            try:
                scalar_file_content = scalar_file.read()
            except UnicodeDecodeError:
                raise NonTextFileError((
                    f"Non-text file {scalar_file_path} found in {dir_path}. "
                    "Only plain text files and folders can be read by "
                    "yamlex. To ignore the file, prefix it with an "
                    f"exclamation mark like so: !{scalar_file_name}."
                ))
            scalar_node: str | FoldedScalarString = scalar_file_content
            if keep_formatting:
                scalar_node = FoldedScalarString(scalar_file_content)

            if scalar_file_name in data:
                raise DuplicateKey(
                    f"Duplicate key found inside {dir_path}: {yaml_file_name}"
                )

            data[scalar_file_name] = scalar_node

    # Recursively traverse directories.
    # Directory's name is considered to be the name of the nested field,
    # if it's not an array. Otherwise, directory name is ignored.
    for sub_dir in all_dirs:
        sub_dir_data = assemble_recursively(
            sub_dir,
            sort_paths=sort_paths,
            dry_run=dry_run,
            remove_comments=remove_comments,
            level=level + 1,
        )

        if sub_dir.name in data:
            raise DuplicateKey(
                f"Duplicate key found inside {dir_path}: {sub_dir.name}"
            )

        data[sub_dir.name] = sub_dir_data

    # Current directory is an array if either
    #   a. There is at least 1 item in it prefixed with a dash
    #   b. There is at least 1 grouper object that is a list
    is_current_dir_array = False
    for k, v in data.items():
        if k.startswith("-"):
            is_current_dir_array = True
            logger.debug((
                f"{indent}Level {dir_path} is an array because an item "
                f"with the dash prefix was found: {k}"
            ))
            break
        if k.startswith("+") and isinstance(v, list):
            is_current_dir_array = True
            logger.debug((
                f"{indent}Level {dir_path} is an array because a grouper "
                f"containing an array was found: {k}"
            ))
            break

    result_as_list: list = []
    result_as_dict: dict = {}
    for k, v in data.items():
        if k == "index" and isinstance(v, list):
            raise IndexFileIsArray((
                f"Directory {dir_path} contains an index file with an "
                "array."
            ))

        if is_current_dir_array:
            if k.startswith("-"):
                result_as_list.append(v)
            elif k.startswith("+") and isinstance(v, list):
                result_as_list.extend(v)
            else:
                # If current dir is an array, then all items in it
                # must be valid array items too.
                plain = k.lstrip("-+")
                raise InvalidItemWithinArrayDirectoryError((
                    f"Directory {dir_path} is an array, but it contains an "
                    f"unexpected value in key {k} (type: {type(v)}). Inside array folders, "
                    f"either the key should start with a dash -{plain} or "
                    f"the key must be a grouper +{plain} that contains "
                    f"another array within itself."
                ))
            
        else:
            if (k.startswith("+") or k == "index") and isinstance(v, dict):
                result_as_dict.update(v)
            else:
                result_as_dict[k] = v

    result = result_as_list if is_current_dir_array else result_as_dict
    logger.debug(f"{indent}Level {dir_path} returned {type(result)}")
    return result



import logging
import warnings
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from yamlex.api.joiner import assemble_recursively
from yamlex.api.util import (
    adjust_root_logger,
    get_default_extension_dir_path,
    get_default_extension_source_dir_path,
    is_manually_created,
    write_file,
    read_version_properties,
    parse_version,
)
from yamlex.api.exceptions import (
    WrongExtensionStructureError,
    EmptyAssembledExtensionError,
    OverwritingManuallyCreatedFileError,
    UnintendedIndexFileWarning,
)
from yamlex.cli.common_flags import (
    no_file_header_flag,
    force_flag,
    verbose_flag,
    quiet_flag,
    dry_run_flag,
    remove_comments_flag,
    line_length_option,
)


logger = logging.getLogger(__name__)


def join(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help="Path to the directory where individual source component files are stored.",
            show_default="source or src/source",
            dir_okay=True,
            file_okay=False,
            exists=True,
            readable=True,
        )
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help="Path to the target extension.yaml file that will be assembled from parts.",
            show_default="extension/extension.yaml or src/extension/extension.yaml",
            dir_okay=False,
            file_okay=True,
        )
    ] = None,
    dev: Annotated[
        bool,
        typer.Option(
            "--dev",
            "-d",
            help="Add the 'custom:' prefix and embed version into extension.yaml.",
        ),
    ] = False,
    keep_formating: Annotated[
        bool,
        typer.Option(
            "--keep",
            "-k",
            help="Keep formatting and indentation when embedding non-yaml files.",
        ),
    ] = True,
    version: Annotated[
        Optional[str],
        typer.Option(
            "--version",
            "-v",
            help="Explicitly set the version in the extension.yaml.",
            show_default=False,
        ),
    ] = None,
    sort_paths: Annotated[
        bool,
        typer.Option(
            "--sort-paths",
            help="Sort paths alphabetically when traversing source directory before join.",
        ),
    ] = False,
    line_length: line_length_option = None,
    dry_run: dry_run_flag = False,
    remove_comments: remove_comments_flag = False,
    no_file_header: no_file_header_flag = False,
    force: force_flag = False,
    verbose: verbose_flag = False,
    quiet: quiet_flag = False,
) -> None:
    """
    Join individual components into a single [i]extension.yaml[/i] file.

    Assembles all files from the --source directory in a hierarchical order.
    As if the folder structure of the --source directory represents a YAML
    structure.

    [b]Overwriting existing extension.yaml (--no-file-header and --force)[/b]

    Yamlex tries to be cautious not to accidentally overwrite a manually
    created [i]extension.yaml[/i]. If that file contains the "Generated with
    yamlex" header in it, then yamlex overwrites it without hesitation.
    However, when [i]extension.yaml[/i] does not contain that line, yamlex
    does not overwrite it. You can alter this behaviour using --force flag.
    
    When yamlex generates the [i]extension.yaml[/i] from parts, it adds the
    same comment at the top: # Generated by yamlex
    If you would like to disable this behaviour, use the
    --no-file-header flag.

    [b]Development mode[/b]

    When you add the --dev flag, yamlex will add the "custom:" prefix to the
    name of your extension and will put an explicit version into the final
    [i]extension.yaml[/i].
    """
    adjust_root_logger(verbose, quiet)

    source = source or get_default_extension_source_dir_path()
    logger.debug(f"Source files directory: {source}")

    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter("always")
        extension = assemble_recursively(
            source,
            keep_formatting=keep_formating,
            sort_paths=sort_paths,
            dry_run=dry_run,
            remove_comments=remove_comments,
        )

        index_file_paths: list[str] = []
        for w in caught_warnings:
            if w.category is UnintendedIndexFileWarning:
                index_file_paths.append(w.message.args[0])

        # Explicitly disabled until we bump major version. For now we
        # support index.yaml files as usual.
        # if dry_run and index_file_paths:
        #     print(f"Warning! index.yaml files found.")
        #     for i in index_file_paths:
        #         print(i)
        #     print((
        #         "Newer versions of Yamlex no longer use index files as "
        #         "groupers. If you intend to have a field called 'index', "
        #         "ignore this warning. Otherwise, prefix the index file with "
        #         f"a grouper sign like so: +index.yaml. This warning will be "
        #         "removed in the later versions of Yamlex."
        #     ))

    # An assembled extension cannot be an array.
    if isinstance(extension, list):
        raise WrongExtensionStructureError((
            "Error! Invalid source directory structure. "
            "Assembled extension is an array while object is expected."
        ))

    # An assembled extension cannot be empty
    if not extension:
        raise EmptyAssembledExtensionError(
            "Error! Failed to assemble the extension. The result is empty."
        )

    # Figure out the current version
    yaml_version = extension.get("version")

    if dev:
        name = extension.get("name")
        if isinstance(name, str) and not name.startswith("custom:"):
            name = f"custom:{name}"
            extension["name"] = name
            logger.info(f"Dev mode extension name: {name}")

        # If new version is not explicitly specified
        if not version:
            current_version = parse_version(yaml_version)
            if not current_version:
                current_version = read_version_properties("version.properties")
                logger.info(f"Version found in version.properties: {current_version}")

        # Update version, if new explicit version was specified
        if not version:
            version = current_version

        # Explicitly add version to generated extension.yaml
        if yaml_version != version:
            extension["version"] = version
            logger.info(f"Explicit version in dev mode: {version}")

    target = target or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"Target file: {target}")

    # Check if the target file exists and was created manually
    if is_manually_created(target) and not force:
        raise OverwritingManuallyCreatedFileError(
            f"The {target} file was created manually. Use --force to overwrite it."
        )

    # Write to output file
    should_add_file_header = not no_file_header
    write_file(
        target,
        extension,
        add_file_header=should_add_file_header,
        line_length=line_length,
        dry_run=dry_run,
    )



class YamlexError(Exception):
    """Base class for all Yamlex errors."""
    code = 10


class YamlexWarning(UserWarning):
    """Base class for all Yamlex warnings."""
    pass


class InvalidItemWithinArrayDirectoryError(YamlexError):
    code = 11


class UnintendedIndexFileWarning(UserWarning):
    code = 12


class NonTextFileError(YamlexError):
    code = 13


class FailedToParseYamlError(YamlexError):
    code = 14


class WrongExtensionStructureError(YamlexError):
    code = 15


class EmptyAssembledExtensionError(YamlexError):
    code = 16


class OverwritingManuallyCreatedFileError(YamlexError):
    code = 17


class IndexFileIsArray(YamlexError):
    code = 18


class InvalidPath(YamlexError):
    code = 19


class NoValidVersionNumber(YamlexError):
    code = 20


class MissingExtensionSchema(YamlexError):
    code = 21


class ExtensionDirNotFound(YamlexError):
    code = 22


class FailedToReadVersionFile(YamlexError):
    code = 23


class FailedToWriteVersionFile(YamlexError):
    code = 24


class DuplicateKey(YamlexError):
    code = 25



import importlib.metadata
import re
from pathlib import Path


def get_version() -> None:
    version = "0.0.0"
    try:
        metadata = importlib.metadata.metadata("yamlex")
        version = metadata["Version"]
    except importlib.metadata.PackageNotFoundError:
        pyproject_file_path = Path(__file__).parent.parent / "pyproject.toml"
        if pyproject_file_path.exists() and pyproject_file_path.is_file():
            with open(pyproject_file_path) as f:
                content = f.read()
                match_ = re.search(r"version\s*=\s*(.+)", content)
                if match_:
                    version_group = match_.group(1)
                    version = version_group.strip("\"")

    return version



import json
import logging
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from yamlex.api.mapper import (
    validate_json_schemas_dir,
    map_schema_to_sources,
    read_vscode_settings_json_file,
    extract_definitions_into_standalone_schemas,
)
from yamlex.api.util import (
    adjust_root_logger,
    get_default_extension_dir_path,
    get_default_extension_source_dir_path,
)

from yamlex.cli.common_flags import (
    verbose_flag,
    quiet_flag,
)


logger = logging.getLogger(__name__)


def map(
    settings: Annotated[
        Path,
        typer.Argument(
            help="Path to the VS Code settings.json file.",
        ),
    ] = Path(".vscode/settings.json"),
    schema: Annotated[
        Path,
        typer.Option(
            "--json",
            "-j",
            help="Path to directory with valid extensions JSON schema files.",
            exists=True,
            readable=True,
            dir_okay=True,
            file_okay=False,
        ),
    ] = Path("schema"),
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help="Path to directory where YAML source files will be stored.",
            show_default="source or src/source",
            dir_okay=True,
            file_okay=False,
        )
    ] = None,
    root: Annotated[
        Path,
        typer.Option(
            "--root",
            "-r",
            help="Root directory relative to which the paths in settings file will be mapped.",
            dir_okay=True,
            file_okay=False,
        )
    ] = Path("."),
    extension_yaml: Annotated[
        Optional[Path],
        typer.Option(
            "--extension-yaml",
            "-e",
            help="Path to output extension.yaml file.",
            show_default="extension/extension.yaml or src/extension/extension.yaml",
            dir_okay=False,
            file_okay=True,
        )
    ] = None,
    verbose: verbose_flag = False,
    quiet: quiet_flag = False,
) -> None:
    """
    Map JSON schema to YAML files in VS Code settings.
    """
    adjust_root_logger(verbose, quiet)
    logger.debug(f"JSON schema files directory: {schema}")
    logger.debug(f"VS Code settings.json file: {settings}")
    logger.debug(f"Root dir: {root}")

    source = source or get_default_extension_source_dir_path()
    logger.debug(f"Extension YAML source files directory: {source}")

    extension_yaml = extension_yaml or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"extension.yaml file: {extension_yaml}")

    # Make sure schemas directory contains extension.schema.json
    validate_json_schemas_dir(schema)

    # Make schema more granular by extracting embedded definitions into
    # separate schema files.
    extract_definitions_into_standalone_schemas(schema)

    # Update yaml.schemas mapping in settings.json
    vscode_settings = read_vscode_settings_json_file(settings)

    # Delete existing extension.yaml mapping
    mapping: dict = vscode_settings.get("yaml.schemas", {})
    for k, v in mapping.items():
        if "extension.yaml" in str(v):
            del mapping[k]
            break
    
    # Update the mapping
    mapping_update = map_schema_to_sources(schema, source, root, extension_yaml)
    mapping.update(mapping_update)

    # Write the updated settings.json file
    vscode_settings["yaml.schemas"] = mapping
    with open(settings, "w") as f:
        vscode_settings_file_text = json.dumps(vscode_settings, indent=2)
        f.write(vscode_settings_file_text)
        logger.info(f"Updated {settings} with new YAML schema mappings.")